/*

ItemMotionHandler
ItemMotionInterface

K			LightWave 6.0
R|[lg	CAEg
wb_[		lwmotion.h

#include <lwmotion.h>

LWITEMMOTION_HCLASS
LWITEMMOTION_ICLASS
LWITEMMOTION_VERSION

*/


#include "main.h"
#include "LW_CVector.h"



static GlobalFunc		*g_global = NULL;
static EZMuscleGlobal	*g_muscleGlobal = NULL;


//CX^X֐
XCALL_(LWInstance)		Create(void *priv, void *context, LWError *error);
XCALL_(void)			Destroy(LWInstance inst);
XCALL_(LWError)			Copy(LWInstance dest, LWInstance source);
XCALL_(LWError)			Load(LWInstance inst, const LWLoadState *ls);
XCALL_(LWError)			Save(LWInstance inst, const LWSaveState *ss);
XCALL_(const char *)	Descln(LWInstance inst);

//ACenh[̊g
XCALL_(const LWItemID *)	UseItems(LWInstance inst);
XCALL_(void)				ChangeID(LWInstance inst, const LWItemID *id);

//nh[֐
XCALL_(void)			Evaluate(LWInstance inst, const LWItemMotionAccess *access);
XCALL_(unsigned int)	Flags(LWInstance inst);


//----------------------------------------------------------------------------------------------------
//	nh[ANeBx[V֐
//----------------------------------------------------------------------------------------------------
XCALL_(int)EZMuscleActivate(long version, GlobalFunc *global, LWItemMotionHandler *local, void *serverData)
{
	if( LWITEMMOTION_VERSION != version )
		return AFUNC_BADVERSION;

	g_global = global;

	local->inst->create		= Create;
	local->inst->destroy	= Destroy;
	local->inst->copy		= Copy;
	local->inst->load		= Load;
	local->inst->save		= Save;
	local->inst->descln		= Descln;
	local->item->useItems	= UseItems;
	local->item->changeID	= ChangeID;
	local->evaluate			= Evaluate;
	local->flags			= Flags;

	return AFUNC_OK;
}



XCALL_(LWInstance) Create(void *priv, void *context, LWError *error)
{
	EZMuscleData	*data;
	MuscleParam		param;
	LWGlobalPool	*memfunc = (LWGlobalPool*)g_global( LWGLOBALPOOL_GLOBAL, GFUSE_TRANSIENT );

	data = new EZMuscleData;
	
	data->self = (LWItemID)context;
	data->global = g_global;
	data->iinfo = (LWItemInfo*)g_global( LWITEMINFO_GLOBAL, GFUSE_TRANSIENT );
	strcpy(data->desc, SN_EZMUSCLE);
	data->binfo = (LWBoneInfo*)g_global( LWBONEINFO_GLOBAL, GFUSE_TRANSIENT );

	param.items[0] = data->self;
	param.items[1] = NULL;
	param.items[2] = NULL;
	param.holdVolume = false;
	param.afterIK = true;
	memset(param.refPos, 0, sizeof(double)*3);
	param.restLength = data->binfo->restLength(param.items[0]);
	param.refresh = false;

	int				index;
	EZMuscleGlobal	**muscleGlobal = (EZMuscleGlobal**)(memfunc->find(SN_EZMUSCLE));
	if( muscleGlobal ){
		if( NULL == *muscleGlobal ){
			*muscleGlobal = new EZMuscleGlobal;
			g_muscleGlobal = *muscleGlobal;
			(*muscleGlobal)->reference = 0;
		}
		(*muscleGlobal)->reference++;
		index = getMuscleGlobalIndex(*muscleGlobal, param.items[0]);
		if( -1 == index ){
			(*muscleGlobal)->params.push_back(param);
			data->param = (*muscleGlobal)->params.begin() + ((*muscleGlobal)->params.size()-1);
		}else{
			data->param = ((*muscleGlobal)->params.begin() + index);
		}
	}else{
		muscleGlobal = (EZMuscleGlobal**)(memfunc->create(SN_EZMUSCLE, sizeof(EZMuscleGlobal**)));
		*muscleGlobal = new EZMuscleGlobal;
		g_muscleGlobal = *muscleGlobal;
		(*muscleGlobal)->reference = 1;
		(*muscleGlobal)->params.push_back(param);
		data->param = (*muscleGlobal)->params.begin();
	}

	return data;
}

XCALL_(void) Destroy(LWInstance inst)
{
	EZMuscleData	*data = (EZMuscleData*)inst;
	LWGlobalPool	*memfunc = (LWGlobalPool*)data->global( LWGLOBALPOOL_GLOBAL, GFUSE_TRANSIENT );

	EZMuscleGlobal	**muscleGlobal = (EZMuscleGlobal**)(memfunc->find(SN_EZMUSCLE));
	if( muscleGlobal ){
		(*muscleGlobal)->reference--;
		if( 0 == (*muscleGlobal)->reference ){
			delete *muscleGlobal;
			*muscleGlobal = NULL;
			g_muscleGlobal = NULL;
		}
	}else if( g_muscleGlobal ){
		delete g_muscleGlobal;
		g_muscleGlobal = NULL;
	}
}


XCALL_(LWError) Copy(LWInstance dest, LWInstance source)
{
	EZMuscleData	*from, *to;

	from = (EZMuscleData*)source;
	to = (EZMuscleData*)dest;

	to->global = from->global;
	strcpy(to->desc, from->desc);
	to->iinfo = from->iinfo;
	to->binfo = from->binfo;

	to->param->afterIK = from->param->afterIK;
	to->param->holdVolume = from->param->holdVolume;
	to->param->items[1] = from->param->items[1];

	refreshSettings(to, to->self, g_muscleGlobal);

	return NULL;
}


XCALL_(LWError) Load(LWInstance inst, const LWLoadState *ls)
{
	EZMuscleData	*data = (EZMuscleData*)inst;
	unsigned long	u4;
	LWFVector		fvec;
	unsigned char	u1;

	refreshSettings(data, data->self, g_muscleGlobal);

	LWLOAD_U4(ls, &u4, 1);
	data->param->items[0] = (LWItemID)u4;
	LWLOAD_U4(ls, &u4, 1);
	data->param->items[1] = (LWItemID)u4;
	LWLOAD_U1(ls, &u1, 1);
	data->param->holdVolume = u1 ? true : false;
	LWLOAD_U1(ls, &u1, 1);
	data->param->afterIK = u1 ? true : false;
	LWLOAD_FP(ls, fvec, 1);
	data->param->restLength = (double)fvec[0];
	LWLOAD_FP(ls, fvec, 3);
	for(int i=0; i<3; i++) data->param->refPos[i] = (double)fvec[i];
	for(i=0; i<g_muscleGlobal->params.size(); i++) g_muscleGlobal->params[i].refresh = true;

	data->global = g_global;
	strcpy(data->desc, SN_EZMUSCLE);
	data->iinfo = (LWItemInfo*)g_global( LWITEMINFO_GLOBAL, GFUSE_TRANSIENT );
	data->binfo = (LWBoneInfo*)g_global( LWBONEINFO_GLOBAL, GFUSE_TRANSIENT );

	return NULL;
}


XCALL_(LWError) Save(LWInstance inst, const LWSaveState *ss)
{
	EZMuscleData	*data = (EZMuscleData*)inst;
	unsigned long	u4;
	LWFVector		fvec;
	unsigned char	u1;

	u4 = (unsigned long)data->param->items[0];
	LWSAVE_U4(ss, &u4, 1);
	u4 = (unsigned long)data->param->items[1];
	LWSAVE_U4(ss, &u4, 1);
	u1 = data->param->holdVolume ? 1 : 0;
	LWSAVE_U1(ss, &u1, 1);
	u1 = data->param->afterIK ? 1 : 0;
	LWSAVE_U1(ss, &u1, 1);
	fvec[0] = (float)data->param->restLength;
	LWSAVE_FP(ss, fvec, 1);
	for(int i=0; i<3; i++) fvec[i] = (float)data->param->refPos[i];
	LWSAVE_FP(ss, fvec, 3);

	return NULL;
}


XCALL_(const char *) Descln(LWInstance inst)
{
	EZMuscleData	*data = (EZMuscleData*)inst;
	const char		*name = data->iinfo->name(data->param->items[1]);

	if( !name ){
		strcpy(data->desc, SN_EZMUSCLE);
	}else{
		sprintf(data->desc, "%s (%s)", SN_EZMUSCLE, name);
	}
	return (const char *)data->desc;
}


XCALL_(const LWItemID *) UseItems(LWInstance inst)
{
	EZMuscleData	*data = (EZMuscleData*)inst;

	return data->param->items;
}


XCALL_(void) ChangeID(LWInstance inst, const LWItemID *id)
{
	EZMuscleData	*data = (EZMuscleData*)inst;

	for(;*id != NULL; id+=2){
		if( id[0] == data->param->items[0] ){
			data->param->items[0] = id[1];
		}
		if( id[0] == data->param->items[1] ){
			data->param->items[1] = id[1];
		}
	}
}



//[J[hϊ
LW_CMatrix_4x4 LtoW(LWItemID id, const LWItemInfo *iinfo, double time)
{
	LWItemID		parent;
	LW_CMatrix_4x4	mat;
	LWDVector		vec;
	
	mat.init();

	iinfo->param(id, LWIP_ROTATION, time, vec);
	mat.rotateHPB(vec[0], vec[1], vec[2]);

	iinfo->param(id, LWIP_POSITION, time, vec);
	mat.translate(vec[0], vec[1], vec[2]);
	
	parent = iinfo->parent(id);
	if( parent ){
		mat = mat * LtoW(parent, iinfo, time);
	}
	
	return mat;
}

//[h[Jϊ
LW_CMatrix_4x4 WtoL(LWItemID id, const LWItemInfo *iinfo, double time)
{
	LWItemID		parent;
	LW_CMatrix_4x4	mat;
	LWDVector		vec;
	int				i;
	
	mat.init();

	iinfo->param(id, LWIP_POSITION, time, vec);
	for(i=0; i<3; i++) vec[i] *= -1;
	mat.translate(vec[0], vec[1], vec[2]);
	
	iinfo->param(id, LWIP_ROTATION, time, vec);
	for(i=0; i<3; i++) vec[i] *= -1;
	mat.rotateY(vec[0]);
	mat.rotateX(vec[1]);
	mat.rotateZ(vec[2]);

	parent = iinfo->parent(id);
	if( parent ){
		mat = WtoL(parent, iinfo, time) * mat;
	}
	
	return mat;
}


XCALL_(void) Evaluate(LWInstance inst, const LWItemMotionAccess *access)
{
	EZMuscleData	*data = (EZMuscleData*)inst;
	LWDVector		rot, scl, tip;
	double			length;
	LW_CMatrix_4x4	mat;
	
	if( data->param->refresh ) refreshSettings(data, access->item, g_muscleGlobal);

	data->param->items[0] = access->item;
	if( NULL == data->param->items[1] ) return;

	memcpy(tip, data->param->refPos, sizeof(double)*3);
	//[J(reference)[h
	mat = LtoW(data->param->items[1], data->iinfo, access->time);
	mat.Transform(tip);

	//[h[J(self)
	mat = WtoL(data->param->items[0], data->iinfo, access->time);
	mat.Transform(tip);

	access->getParam(LWIP_ROTATION, access->time, rot);
	mat.init();
	mat.rotateHPB(rot[0], rot[1], rot[2]);
	mat.Transform(tip);
	Angle(tip, rot);
	length = Distance(tip);

	access->getParam(LWIP_SCALING, access->time, scl);
	scl[2] = length / data->param->restLength;
	if( data->param->holdVolume ){
		scl[0] = scl[1] = sqrt(1/scl[2]);
	}

	access->setParam(LWIP_ROTATION, rot);
	access->setParam(LWIP_SCALING, scl);
}


XCALL_(unsigned int) Flags(LWInstance inst)
{
	EZMuscleData	*data = (EZMuscleData*)inst;

	if( data->param->afterIK ){
		return LWIMF_AFTERIK;
	}else{
		return 0;
	}
}



int getMuscleGlobalIndex(EZMuscleGlobal *mg, const LWItemID id)
{
	for(int i=0; i<mg->params.size(); i++){
		if( id == mg->params[i].items[0] ) return i;
	}
	return -1;
}

EZMuscleGlobal * getMuscleGlobal(LWGlobalPool *memfunc)
{
	EZMuscleGlobal **mg = (EZMuscleGlobal**)(memfunc->find(SN_EZMUSCLE));
	if( !mg ){
		mg = (EZMuscleGlobal**)(memfunc->create(SN_EZMUSCLE, sizeof(EZMuscleGlobal**)));
		*mg = new EZMuscleGlobal;
		g_muscleGlobal = *mg;
		(*mg)->reference = 0;
	}
	return *mg;
}



