#ifndef	__MAIN_H__
#define __MAIN_H__

#ifndef	NULL
#ifdef _cplusplus
#define	NULL	0
#endif
#endif

#ifndef	MAX_PATH
#define	MAX_PATH	260
#endif

#ifndef	PI
#define	PI		3.14159265358979
#endif

#ifndef	HALF_PI
#define	HALF_PI		1.57079632679489
#endif


#include <vector>

#include <lwserver.h>
#include <lwhost.h>
#include <lwcmdseq.h>
#include <lwgeneric.h>
#include <lwrender.h>
#include <lwmotion.h>

#include "LW_CMatrix.h"


#define	SN_SETMUSCLE	"SetMuscle"
#define	SN_RDMUSCLE		"ReadMuscle"
#define	SN_EZMUSCLE		"EZMuscle"



#define LWPTAG_COLOR	LWID_('C', 'O', 'L', 'R')
#define LWPTAG_MSLTIP	LWID_('M', 'T', 'I', 'P')
#define	LWPTAG_MSLFLAG	LWID_('M', 'F', 'L', 'G')




typedef struct {
	LWPolID		id;
	int			holdVol;
	int			afterIK;
	LWPolID		tipParent;
}SeMuscleParam;

typedef struct {
	GlobalFunc					*global;
	MeshEditOp					*edit;
	std::vector<LWPolID>		AllBones;
	std::vector<SeMuscleParam>	SelBones;
}SetMuscleData;


typedef struct {
	LWItemID	items[3];
	LWDVector	refPos;
	bool		holdVolume;
	bool		afterIK;
	double		restLength;
	bool		refresh;
}MuscleParam;


typedef struct {
	const char	*name;
	int			holdVol;
	int			afterIK;
	const char	*tipParent;
	int			color;
}ReadMuscleParam;

typedef struct {
	GlobalFunc						*global;
	LWMeshInfo						*mesh;
	std::vector<ReadMuscleParam>	skels;
}ReadMuscleData;


typedef struct {
	LWItemID							self;
	std::vector<MuscleParam>::iterator	param;
	GlobalFunc							*global;
	char								desc[MAX_PATH];
	LWItemInfo							*iinfo;
	LWBoneInfo							*binfo;
}EZMuscleData;


typedef struct {
	std::vector<MuscleParam>	params;
	int							reference;
}EZMuscleGlobal;


XCALL_(int)SetMuscleActivate(long version, GlobalFunc *global, LWModCommand *local, void *serverData);
XCALL_(int)ReadMuscleActivate(long version, GlobalFunc *global, LWLayoutGeneric *local, void *serverData);
XCALL_(int)EZMuscleActivate(long version, GlobalFunc *global, LWItemMotionHandler *local, void *serverData);
XCALL_(int)EZMuscleInterface(long version, GlobalFunc *global, LWInterface *local, void *serverData);

int				SetMusclePanel(SetMuscleData *data, GlobalFunc *global);
LW_CMatrix_4x4	LtoW(LWItemID id, const LWItemInfo *iinfo, double time);
LW_CMatrix_4x4	WtoL(LWItemID id, const LWItemInfo *iinfo, double time);
void			refreshSettings(EZMuscleData *data, LWItemID id, EZMuscleGlobal *mg);
int				getMuscleGlobalIndex(EZMuscleGlobal *mg, const LWItemID id);
EZMuscleGlobal*	getMuscleGlobal(LWGlobalPool *memfunc);


#endif  //__MAIN_H__


